@@ -41,9 +41,12 @@
  * without this exception; this exception also makes it possible to
  * release a modified version which carries forward this exception.
  *
- */
+#*/
+
 
 #include "timblapi.h"
+#include "timbl/GetOptClass.h"
+#include "timbl/Instance.h"
 #include "docstrings.h"
 
 #include <unistd.h>
@@ -56,11 +59,12 @@
 
 #include <boost/utility.hpp>
 #include <boost/python.hpp>
+
 using namespace boost::python;
 
 
 tuple TimblApiWrapper::classify(const std::string& line)
-{
+{ 
 	std::string cls;
 	bool result = Classify(line, cls);
 	return make_tuple(result, cls);
@@ -76,15 +80,70 @@ tuple TimblApiWrapper::classify2(const std::string& line)
 }
 
 
-tuple TimblApiWrapper::classify3(const std::string& line)
+tuple TimblApiWrapper::classify3(const std::string& line, const unsigned char requireddepth)
 {
 	std::string cls;
-	std::string distrib;
 	double distance;
-	bool result = Classify(line, cls, distrib, distance);
-	return make_tuple(result, cls, distrib, distance);
+    const Timbl::ValueDistribution * distrib; 
+    const Timbl::TargetValue * result  = Classify(line, distrib , distance);
+    if (result != NULL) {
+        if ((requireddepth > 0) && (matchDepth() < requireddepth)) {
+            return make_tuple(true, "", python::dict(), 999999);
+        } else {
+            const std::string cls = result->Name();
+            return make_tuple(true, cls, dist2dict(distrib), distance);
+        }
+    } else {
+        return make_tuple(false,"",python::dict(),999999);
+    }
 }
 
+tuple TimblApiWrapper::classify3safe(const std::string& line, const unsigned char requireddepth)
+{
+    PyThreadState * m_thread_state = PyEval_SaveThread(); //release GIL
+    
+
+    
+    pthread_t thisthread = pthread_self();
+    Timbl::TimblExperiment * clonedexp;
+    pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER; //global lock
+    pthread_mutex_lock(&lock);
+    if (experimentpool.find(thisthread) != experimentpool.end()) {
+        clonedexp = experimentpool[thisthread];
+        pthread_mutex_unlock(&lock);
+    } else {
+        pthread_mutex_unlock(&lock);
+        clonedexp = detachedexp->clone();
+        *clonedexp = *detachedexp; //ugly but needed
+        if ( detachedexp->getOptParams() ){
+            clonedexp->setOptParams( detachedexp->getOptParams()->Clone(0) );
+        }
+        pthread_mutex_lock(&lock);
+        experimentpool[thisthread] = clonedexp;
+        pthread_mutex_unlock(&lock);
+    }
+
+    const Timbl::ValueDistribution * distrib; 
+    double distance;
+    const Timbl::TargetValue * result = clonedexp->Classify(line, distrib,distance);
+    if (result != NULL) {
+        if ((requireddepth > 0) && (clonedexp->matchDepth() < requireddepth)) {
+            PyEval_RestoreThread(m_thread_state);
+            m_thread_state = NULL;
+            return make_tuple(true, "", python::dict(), 999999);
+        } else {
+            const std::string cls = result->Name();
+            //const std::string diststring = distrib->DistToString();
+            PyEval_RestoreThread(m_thread_state);
+            m_thread_state = NULL;
+            return make_tuple(true, cls, dist2dict(distrib), distance);
+        }
+    } else {
+        PyEval_RestoreThread(m_thread_state);
+        m_thread_state = NULL; 
+        return make_tuple(false,"",python::dict(),999999);
+    }
+}
 
 std::string TimblApiWrapper::bestNeighbours()
 {
@@ -136,6 +195,11 @@ std::string TimblApiWrapper::settings()
 	return buf.str();
 }
 
+
+void TimblApiWrapper::initthreading() {
+    initExperiment();
+    detachedexp = grabAndDisconnectExp();
+}
 	
 bool TimblApiWrapper::showSettings(object& stream)
 {
@@ -150,6 +214,25 @@ bool TimblApiWrapper::showSettings(object& stream)
 	return ShowSettings(out);
 }
 
+
+python::dict TimblApiWrapper::dist2dict(const Timbl::ValueDistribution * distribution, double minf) const {
+    python::dict result;
+
+    size_t freq;
+
+    Timbl::ValueDistribution::VDlist::const_iterator it = distribution->begin();
+    while ( it != distribution->end() ){
+        Timbl::Vfield *f = it->second;
+        freq = f->Freq();
+        if ( freq >= minf ){
+            result[f->Value()->Name()] = freq;
+        }
+        ++it;
+    }
+
+    return result;
+}
+
 /*std::string TimblApiWrapper::weights()
 {
 	std::ostringstream buf;
@@ -204,6 +287,9 @@ BOOST_PYTHON_MODULE(timblapi)
 		.def("classify", &TimblApiWrapper::classify, CLASSIFY_DOC)
 		.def("classify2", &TimblApiWrapper::classify2, CLASSIFY2_DOC)
 		.def("classify3", &TimblApiWrapper::classify3, CLASSIFY3_DOC)
+		.def("classify3safe", &TimblApiWrapper::classify3safe, CLASSIFY3SAFE_DOC)
+
+		.def("initthreading", &TimblApiWrapper::initthreading, INITTHREADING_DOC)
 
 		.def("showBestNeighbours", &TimblApiWrapper::showBestNeighbours,
 				 SHOWBESTNEIGHBOURS_DOC)
@@ -261,3 +347,6 @@ BOOST_PYTHON_MODULE(timblapi)
 	
 	//def("to_string", to_string);
 }
+
+
+
